@@ -19,7 +19,7 @@ from osmium.replication.server import ReplicationServer
 from shapely.geometry import box
 from tqdm import tqdm
 
-from osmsg.utils import verify_me_osm
+from osmsg.utils import create_charts, create_profile_link, verify_me_osm
 
 from .changefiles import (
     get_prev_hour,
@@ -58,28 +58,35 @@ def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
         if changeset not in users_temp[user]["changesets"]:
             users_temp[user]["changesets"].append(changeset)
         users[user]["changesets"] = len(users_temp[user]["changesets"])
-        if hashtags:
-            for ch in hashtag_changesets[changeset]["countries"]:
-                if ch not in users[user]["countries"]:
-                    users[user]["countries"].append(ch)
-            for ch in hashtag_changesets[changeset]["hashtags"]:
-                if ch not in users[user]["hashtags"]:
-                    users[user]["hashtags"].append(ch)
+        if hashtags or changeset:
+            try:
+                for ch in hashtag_changesets[changeset]["countries"]:
+                    if ch not in users[user]["countries"]:
+                        users[user]["countries"].append(ch)
+                for ch in hashtag_changesets[changeset]["hashtags"]:
+                    if ch not in users[user]["hashtags"]:
+                        users[user]["hashtags"].append(ch)
+            except:
+                pass
 
         users[user][osm_type][action] += 1
         if osm_type == "nodes" and tags:
-            users[user]["poi"][action] += 1
-        if wild_tags:
-            for key, value in tags:
-                if key in users[user][f"tags_{action}"]:
-                    users[user][f"tags_{action}"][key] += 1
-                else:
-                    users[user][f"tags_{action}"][key] = 1
+            if action != "delete":
+                users[user]["poi"][action] += 1
+        if all_tags:
+            if tags:
+                for key, value in tags:
+                    if action != "delete":  # we don't need deleted tags
+                        if key in users[user][f"tags_{action}"]:
+                            users[user][f"tags_{action}"][key] += 1
+                        else:
+                            users[user][f"tags_{action}"][key] = 1
 
         if tags_to_collect:
-            for tag in tags_to_collect:
-                if tag in tags:
-                    users[user][tag][action] += 1
+            if action != "delete":  # we don't need deleted tags
+                for tag in tags_to_collect:
+                    if tag in tags:
+                        users[user][tag][action] += 1
         if country:
             for ch in countries_changesets[changeset]:
                 if ch not in users[user]["hashtags"]:
@@ -93,27 +100,32 @@ def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
             "nodes": {"create": 0, "modify": 0, "delete": 0},
             "ways": {"create": 0, "modify": 0, "delete": 0},
             "relations": {"create": 0, "modify": 0, "delete": 0},
-            "poi": {"create": 0, "modify": 0, "delete": 0},  # nodes that has tags
+            "poi": {"create": 0, "modify": 0},  # nodes that has tags
         }
-        if hashtags:
-            users[user]["countries"] = hashtag_changesets[changeset]["countries"]
-            users[user]["hashtags"] = hashtag_changesets[changeset]["hashtags"]
+        if hashtags or changeset:
+            users[user]["countries"] = []
+            users[user]["hashtags"] = []
+
+            try:
+                users[user]["countries"] = hashtag_changesets[changeset]["countries"]
+                users[user]["hashtags"] = hashtag_changesets[changeset]["hashtags"]
+            except:
+                pass
 
         if tags_to_collect:
             for tag in tags_to_collect:
-                users[user][tag] = {"create": 0, "modify": 0, "delete": 0}
+                users[user][tag] = {"create": 0, "modify": 0}
         users_temp[user] = {"changesets": []}
         if changeset not in users_temp[user]["changesets"]:
             users_temp[user]["changesets"].append(changeset)
         users[user]["changesets"] = len(users_temp[user]["changesets"])
         users[user][osm_type][action] = 1
-        if wild_tags:
+        if all_tags:
             users[user]["tags_create"] = {}
             users[user]["tags_modify"] = {}
-            users[user]["tags_delete"] = {}
-
-            for tag, value in tags:
-                users[user][f"tags_{action}"][tag] = 1
+            if tags:
+                for tag, value in tags:
+                    users[user][f"tags_{action}"][tag] = 1
         if tags_to_collect:
             for tag in tags_to_collect:
                 if tag in tags:
@@ -147,43 +159,58 @@ class ChangesetHandler(osmium.SimpleHandler):
 
     def changeset(self, c):
         country_check = False
+        run_hashtag_check_logic = False
+        if changeset and not hashtags:
+            if "comment" in c.tags:
+                run_hashtag_check_logic = True
         if hashtags:
             if "comment" in c.tags:
-                if any(elem.lower() in c.tags["comment"].lower() for elem in hashtags):
-                    if c.id not in hashtag_changesets.keys():
-                        hashtag_changesets[c.id] = {"hashtags": [], "countries": []}
-                    # get bbox
-                    bounds = str(c.bounds)
-                    if "invalid" not in bounds:
-                        bbox_list = bounds.strip("()").split(" ")
-                        minx, miny = bbox_list[0].split("/")
-                        maxx, maxy = bbox_list[1].split("/")
-                        bbox = box(float(minx), float(miny), float(maxx), float(maxy))
-                        # Create a point for the centroid of the bounding box
-                        centroid = bbox.centroid
-                        intersected_rows = countries_df[
-                            countries_df.intersects(centroid)
-                        ]
-                        hashtags_comment = re.findall(r"#[\w-]+", c.tags["comment"])
-                        for hash_tag in hashtags_comment:
-                            if hash_tag not in hashtag_changesets[c.id]["hashtags"]:
-                                hashtag_changesets[c.id]["hashtags"].append(hash_tag)
-                        for i, row in intersected_rows.iterrows():
-                            if row["name"] not in hashtag_changesets[c.id]["countries"]:
-                                if country:
-                                    country_check = True
-                                    if row["name"] == country:
-                                        hashtag_changesets[c.id]["countries"].append(
-                                            row["name"]
-                                        )
-                                else:
+                if exact_lookup:
+                    hashtags_comment = re.findall(r"#[\w-]+", c.tags["comment"])
+                    if any(
+                        elem.lower() in map(str.lower, hashtags_comment)
+                        for elem in hashtags
+                    ):
+                        run_hashtag_check_logic = True
+                elif any(
+                    elem.lower() in c.tags["comment"].lower() for elem in hashtags
+                ):
+                    run_hashtag_check_logic = True
+
+        if run_hashtag_check_logic:
+            if c.id not in hashtag_changesets.keys():
+                hashtag_changesets[c.id] = {"hashtags": [], "countries": []}
+                # get bbox
+                bounds = str(c.bounds)
+                if "invalid" not in bounds:
+                    bbox_list = bounds.strip("()").split(" ")
+                    minx, miny = bbox_list[0].split("/")
+                    maxx, maxy = bbox_list[1].split("/")
+                    bbox = box(float(minx), float(miny), float(maxx), float(maxy))
+                    # Create a point for the centroid of the bounding box
+                    centroid = bbox.centroid
+                    intersected_rows = countries_df[countries_df.intersects(centroid)]
+                    hashtags_comment = re.findall(r"#[\w-]+", c.tags["comment"])
+                    for hash_tag in hashtags_comment:
+                        if hash_tag not in hashtag_changesets[c.id]["hashtags"]:
+                            hashtag_changesets[c.id]["hashtags"].append(hash_tag)
+                    for i, row in intersected_rows.iterrows():
+                        if row["name"] not in hashtag_changesets[c.id]["countries"]:
+                            if country:
+                                country_check = True
+                                if row["name"] == country:
                                     hashtag_changesets[c.id]["countries"].append(
                                         row["name"]
                                     )
+                            else:
+                                hashtag_changesets[c.id]["countries"].append(
+                                    row["name"]
+                                )
 
         if not country_check:  # hash tag not supplied
-            if country:
-                if c.bounds:
+
+            if c.bounds:
+                if country:
                     bounds = str(c.bounds)
                     if "invalid" not in bounds:
                         bbox_list = bounds.strip("()").split(" ")
@@ -198,6 +225,7 @@ class ChangesetHandler(osmium.SimpleHandler):
                             countries_df.intersects(centroid)
                         ]
                         for i, row in intersected_rows.iterrows():
+
                             if row["name"] == country:
                                 if c.id not in countries_changesets.keys():
                                     countries_changesets[c.id] = []
@@ -316,14 +344,14 @@ def get_download_urls_changefiles(
     # Gets sequence id using timestamp we get from osm api using pyosmium tool
     seq = repl.timestamp_to_sequence(start_date)
     # going one step back to cover all changes only if it is not already behind
-    if (
-        scan_early_seq
-        or (start_date - seq_to_timestamp(repl.get_state_url(seq), timezone)).days < 1
-    ):
-        print(
-            f"Initial fetched difference is {(seq_to_timestamp(repl.get_state_url(seq), timezone) - start_date).days} Hence , Reducing Sequence to Cover all Changes"
-        )
-        seq = seq - 1
+    start_seq_time = seq_to_timestamp(repl.get_state_url(seq), timezone)
+    if scan_early_seq or (start_date - start_seq_time).days < 1:
+        if "minute" in base_url:
+            seq = (
+                seq + int((start_date - start_seq_time).total_seconds() / 60)
+            ) - 60  # go 1 hours ahead
+        else:
+            seq = seq - 1
 
     start_seq = seq
     start_seq_url = repl.get_state_url(start_seq)
@@ -345,16 +373,24 @@ def get_download_urls_changefiles(
     server_seq, server_ts = state_info
     server_ts = server_ts.astimezone(dt.timezone.utc)
     last_seq = server_seq
-    last_ts = server_ts
     if end_date:
         end_date_seq = repl.timestamp_to_sequence(end_date)
-        if end_date_seq:
-            if end_date_seq < last_seq:
-                # get one step ahead to cover all
-                last_seq = end_date_seq + 1
-            else:
-                last_seq = end_date_seq
-            last_ts = end_date
+        last_seq = end_date_seq
+        if "minute" in base_url:
+            last_seq = (
+                last_seq
+                + int(
+                    (
+                        seq_to_timestamp(repl.get_state_url(end_date_seq), timezone)
+                        - end_date
+                    ).total_seconds()
+                    / 60
+                )
+            ) + 60  # go 1 hours ahead
+        else:
+            last_seq += 1
+        if last_seq >= server_seq:
+            last_seq = server_seq
 
     print(
         # f"You have supplied {start_date} : {seq} to {last_ts} : {last_seq} . Latest Server Fetched is : {server_seq} & {in_local_timezone(server_ts,timezone)} on {base_url}\n
@@ -399,7 +435,7 @@ def main():
     )
     parser.add_argument(
         "--end_date",
-        help="End date in the format YYYY-MM-DD HH:M:Sz eg:2023-01-28 17:43:09+05:45",
+        help="End date in the format YYYY-MM-DD HH:M:Sz eg: 2023-01-28 17:43:09+05:45",
         default=dt.datetime.now(),
     )
     parser.add_argument(
@@ -441,7 +477,7 @@ def main():
         "--hashtags",
         nargs="+",
         type=str,
-        help="Hashtags Statstics to Collect : List of hashtags , Limited until daily stats for now",
+        help="Hashtags Statistics to Collect : List of hashtags , Limited until daily stats for now , Only lookups if hashtag is contained on the string , not a exact string lookup on beta",
     )
     parser.add_argument(
         "--force",
@@ -469,14 +505,58 @@ def main():
         help="Your public OSM Change Replication URL ",
     )
 
-    parser.add_argument("--extract_last_week", action="store_true", default=False)
-    parser.add_argument("--extract_last_day", action="store_true", default=False)
-    parser.add_argument("--extract_last_month", action="store_true", default=False)
-    parser.add_argument("--extract_last_year", action="store_true", default=False)
-    parser.add_argument("--extract_last_hour", action="store_true", default=False)
+    parser.add_argument(
+        "--last_week",
+        action="store_true",
+        help="Extract stats for last week",
+        default=False,
+    )
+    parser.add_argument(
+        "--last_day",
+        action="store_true",
+        help="Extract Stats for last day",
+        default=False,
+    )
+    parser.add_argument(
+        "--last_month",
+        action="store_true",
+        help="Extract Stats for last Month",
+        default=False,
+    )
+    parser.add_argument(
+        "--last_year",
+        action="store_true",
+        help="Extract stats for last year",
+        default=False,
+    )
+    parser.add_argument(
+        "--last_hour",
+        action="store_true",
+        help="Extract stats for Last hour",
+        default=False,
+    )
+    parser.add_argument(
+        "--charts",
+        action="store_true",
+        help="Exports Summary Charts along with stats",
+        default=False,
+    )
+    parser.add_argument(
+        "--exact_lookup",
+        action="store_true",
+        help="Exact lookup for hashtags to match exact hashtag supllied , without this hashtag search will search for the existence of text on hashtags and comments",
+        default=False,
+    )
 
     parser.add_argument(
-        "--wild_tags",
+        "--changeset",
+        help="Include hashtag and country informations on the stats. It forces script to process changeset replciation , Careful to use this since changeset replication is minutely",
+        action="store_true",
+        default=False,
+    )
+
+    parser.add_argument(
+        "--all_tags",
         action="store_true",
         help="Extract statistics of all of the unique tags and its count",
         default=False,
@@ -493,7 +573,7 @@ def main():
         "--format",
         nargs="+",
         choices=["csv", "json", "excel", "image", "text"],
-        default="json",
+        default="csv",
         help="Stats output format",
     )
     parser.add_argument(
@@ -509,11 +589,11 @@ def main():
 
     if not args.start_date:
         if (
-            args.extract_last_week
-            or args.extract_last_day
-            or args.extract_last_month
-            or args.extract_last_year
-            or args.extract_last_hour
+            args.last_week
+            or args.last_day
+            or args.last_month
+            or args.last_year
+            or args.last_hour
         ):
             pass
         else:
@@ -530,18 +610,30 @@ def main():
         if not countries_df["name"].isin([args.country]).any():
             print("Country doesn't exists : Refer to data/countries_un.csv name column")
             sys.exit()
+    if args.changeset:
+        if args.hashtags or args.country:
+            assert (
+                args.changeset
+            ), "You can not use include changeset meta option along with hashtags/country"
+
     start_time = time.time()
 
     global additional_tags
     global cookies
-    global wild_tags
+    global all_tags
     global hashtags
     global country
-    wild_tags = args.wild_tags
+    global changeset
+    global exact_lookup
+
+    all_tags = args.all_tags
     additional_tags = args.tags
     hashtags = args.hashtags
     country = args.country
     cookies = None
+    changeset = args.changeset
+    exact_lookup = args.exact_lookup
+
     if "geofabrik" in args.url.lower():
         if args.username is None:
             args.username = os.environ.get("OSM_USERNAME")
@@ -555,18 +647,18 @@ def main():
         cookies = auth(args.username, args.password)
     print("Script Started")
 
-    if args.extract_last_hour:
+    if args.last_hour:
         start_date, end_date = get_prev_hour(args.timezone)
 
-    if args.extract_last_year:
+    if args.last_year:
         start_date, end_date = get_prev_year_dates(args.timezone)
 
-    if args.extract_last_month:
+    if args.last_month:
         start_date, end_date = previous_month(args.timezone)
 
-    if args.extract_last_day:
+    if args.last_day:
         start_date, end_date = previous_day(args.timezone)
-    if args.extract_last_week:
+    if args.last_week:
         start_date, end_date = previous_week(args.timezone)
 
     if args.read_from_metadata:
@@ -595,8 +687,9 @@ def main():
             )
             if not args.force:
                 sys.exit()
+    print(f"Supplied start_date: {start_date} and end_date: {end_date}")
 
-    if args.hashtags or args.country:
+    if args.hashtags or args.country or args.changeset:
 
         Changeset = ChangesetToolKit()
         (
@@ -605,7 +698,7 @@ def main():
             changeset_end_seq,
         ) = Changeset.get_download_urls(start_date, end_date)
         print(
-            f"You have supplied start_date as : {start_date} and end_date as : {end_date} , Processing Changeset from {strip_utc(Changeset.sequence_to_timestamp(changeset_start_seq),args.timezone)} to {strip_utc(Changeset.sequence_to_timestamp(changeset_end_seq),args.timezone)}"
+            f"Processing Changeset from {strip_utc(Changeset.sequence_to_timestamp(changeset_start_seq),args.timezone)} to {strip_utc(Changeset.sequence_to_timestamp(changeset_end_seq),args.timezone)}"
         )
 
         temp_path = os.path.join(os.getcwd(), "temp/changesets")
@@ -629,9 +722,9 @@ def main():
             # executor.shutdown(wait=True)
 
         print("Changeset Processing Finished")
-        end_date = strip_utc(
-            Changeset.sequence_to_timestamp(changeset_end_seq), args.timezone
-        )
+        end_seq_timestamp = Changeset.sequence_to_timestamp(changeset_end_seq)
+        if end_date > end_seq_timestamp:
+            end_date = strip_utc(end_seq_timestamp, args.timezone)
 
     print("Changefiles : Generating Download Urls")
     (
@@ -655,6 +748,7 @@ def main():
 
     start_date_utc = start_date.astimezone(dt.timezone.utc)
     end_date_utc = end_date.astimezone(dt.timezone.utc)
+    print(f"Final UTC Date time to filter stats : {start_date_utc} to {end_date_utc}")
     temp_path = os.path.join(os.getcwd(), "temp/changefiles")
     if not os.path.exists(temp_path):
         os.makedirs(temp_path)
@@ -680,7 +774,7 @@ def main():
     shutil.rmtree("temp")
     if len(users) >= 1:
         # print(users)
-        if args.wild_tags:
+        if args.all_tags:
             for user in users:
                 users[user]["tags_create"] = json.dumps(
                     dict(
@@ -700,18 +794,10 @@ def main():
                         )
                     )
                 )
-                users[user]["tags_delete"] = json.dumps(
-                    dict(
-                        sorted(
-                            users[user]["tags_delete"].items(),
-                            key=lambda item: item[1],
-                            reverse=True,
-                        )
-                    )
-                )
+
         df = pd.json_normalize(list(users.values()))
 
-        if hashtags:
+        if hashtags or changeset:
             df["countries"] = df["countries"].apply(lambda x: ",".join(map(str, x)))
 
         df = df.assign(
@@ -732,11 +818,11 @@ def main():
         if args.rows:
             df = df.head(args.rows)
         print(df)
-        if args.wild_tags:
+        if args.all_tags:
             # Get the column names of the DataFrame
             cols = df.columns.tolist()
             # Identify the column names that you want to move
-            cols_to_move = ["tags_create", "tags_modify", "tags_delete"]
+            cols_to_move = ["tags_create", "tags_modify"]
             # Remove the columns to move from the list of column names
             cols = [col for col in cols if col not in cols_to_move]
             # Add the columns to move to the end of the list of column names
@@ -744,28 +830,54 @@ def main():
             # Reindex the DataFrame with the new order of column names
             df = df.reindex(columns=cols)
 
-        if country or hashtags:
+        if country or hashtags or changeset:
             df["hashtags"] = df["hashtags"].apply(lambda x: ",".join(map(str, x)))
             column_to_move = "hashtags"
             df = df.assign(**{column_to_move: df.pop(column_to_move)})
 
-        start_date = in_local_timezone(start_date, args.timezone)
-        end_date = in_local_timezone(end_date, args.timezone)
+        start_date = in_local_timezone(start_date_utc, args.timezone)
+        end_date = in_local_timezone(end_date_utc, args.timezone)
 
         fname = f"{args.name}_{start_date}_{end_date}"
         if args.exclude_date_in_name:
             fname = args.name
-        if "image" in args.format:
+        if "image" in args.format:  ### used for twitter tweet
             # Convert the DataFrame to an image
-            df_img = (
-                (df if args.rows <= 100 else df.head(100))
-                if args.rows
-                else df.head(100)
-            )  # 100 as max rows for image format
+            df_img = df.head(25)
+            # Compute sums of specified columns for the top 20 rows
+            created = df_img[["nodes.create", "ways.create", "relations.create"]].sum(
+                axis=1
+            )
+            modified = df_img[["nodes.modify", "ways.modify", "relations.modify"]].sum(
+                axis=1
+            )
+            deleted = df_img[["nodes.delete", "ways.delete", "relations.delete"]].sum(
+                axis=1
+            )
+
+            # Concatenate original DataFrame and sums DataFrame
+            result_df = pd.concat(
+                [
+                    df_img,
+                    created.rename("Created"),
+                    modified.rename("Modified"),
+                    deleted.rename("Deleted"),
+                ],
+                axis=1,
+            )
+
+            cols_to_export = [
+                "rank",
+                "name",
+                "changesets",
+                "map_changes",
+                "Created",
+                "Modified",
+                "Deleted",
+            ]  # Specify columns to export
+
             dfi.export(
-                df_img.drop(columns=["tags_create", "tags_modify", "tags_delete"])
-                if args.wild_tags
-                else df_img,
+                result_df[cols_to_export],
                 f"{fname}.png",
                 max_cols=-1,
                 max_rows=-1,
@@ -776,9 +888,21 @@ def main():
             #     file.write(json.dumps(users))
             df.to_json(f"{fname}.json", orient="records")
         if "csv" in args.format:
-            df.to_csv(f"{fname}.csv", index=False)
+            # Add the start_date and end_date columns to the DataFrame
+            csv_df = df
+            csv_df["start_date"] = start_date_utc
+            csv_df["end_date"] = end_date_utc
+
+            # Create profile link column
+            csv_df.insert(2, "profile", csv_df["name"].apply(create_profile_link))
+
+            csv_df.to_csv(f"{fname}.csv", index=False)
         if "excel" in args.format:
             df.to_excel(f"{fname}.xlsx", index=False)
+
+        if args.charts:
+            create_charts(df)
+
         if "text" in args.format:
             text_output = df.to_markdown(tablefmt="grid", index=False)
             with open(f"{fname}.txt", "w", encoding="utf-8") as file:
